/*
 * Copyright (C) 2012 by CERN/IT/GT/DMS. All rights reserved.
 *
 * Author: Fabrizio Furano, David Smith <David.Smith@cern.ch>
*/

#include <XrdDPM/XrdDPMRedirAcc.hh>
#include <XrdDPM/XrdDPMTrace.hh>
#include <XrdOuc/XrdOucEnv.hh>
#include <XrdOuc/XrdOucString.hh>
#include <XrdSys/XrdSysPthread.hh>
#include <XrdSec/XrdSecEntity.hh>
#include <XrdSys/XrdSysPlugin.hh>
#include <XrdDPM/XrdDPMCommon.hh>
#include <XrdVersion.hh>

#include <vector>

// globals

#if XrdVNUMBER >= 30300
XrdVERSIONINFO(XrdAccAuthorizeObject,XrdDPMRedirAcc)
#endif

namespace DpmRedirAcc {
	XrdSysError	Say(0,"dpmrediracc_");
	XrdOucTrace	Trace(&Say);
	XrdAccAuthorize*	tokAuthorization(0);
}

using namespace DpmRedirAcc;

static DpmRedirConfigOptions *RedirConfig = 0;

/******************************************************************************/
/*                       X r d D P M R e d i r A c c                          */
/******************************************************************************/

XrdDPMRedirAcc::XrdDPMRedirAcc(const char *cfn,int ForceSecondary) :
		ForceSecondary(ForceSecondary) {
	if (DpmCommonConfigProc(Say, cfn, CommonConfig)) {
		throw dmlite::DmException(DMLITE_CFGERR(EINVAL),
			"problem with (common) configuration");
	}
	Trace.What = CommonConfig.OfsTraceLevel;
}

XrdDPMRedirAcc::~XrdDPMRedirAcc() {}

XrdAccPrivs XrdDPMRedirAcc::Access(const XrdSecEntity *Entity,
		const char *path,
		const Access_Operation oper,
		XrdOucEnv *Env)
{
	EPNAME("Access");

	if (!RedirConfig) {
		Say.Emsg("Access","Common redirector configuration not found");
		return XrdAccPriv_None;
	}

	if (!ForceSecondary && oper == AOP_Stat) {
		// may be a discovery stat triggered by a meta-manager.
		// If it's a normal stat the veto & secondary will be
		// called from DPMFinder Locate
		DEBUG("Passing stat directly");
		return XrdAccPriv_Lookup;
	}

	if (ForceSecondary || DpmIdentity::usesPresetID(Env, Entity)) {
		DEBUG("Should use fixed id, proto="
			<< (Entity ? Entity->prot : "[none]"));
		if (tokAuthorization == 0) {
			TRACE(MOST,"Use of fixed id needs a secondary authorization "
				"library to be configured. Denying");
			return XrdAccPriv_None;
		}
		XrdAccPrivs ret_privs = tokAuthorization->Access(Entity, path, oper, Env);
		if (ret_privs == XrdAccPriv_None) {
			return ret_privs;
		}
		bool restr = 1;
		XrdOucString fp;
		try {
			fp  =
				CanonicalisePath(
			SafeCStr(TranslatePath(*RedirConfig, path)), 1);
		} catch (dmlite::DmException &e) {
	                XrdOucString err = "Name error: ";
        	        err += XrdOucString(path)+" ("+DmExStrerror(e)+")";
			DEBUG(err);
			return XrdAccPriv_None;
		} catch(...) {
			Say.Emsg("Access","Unexpected exception");
			return XrdAccPriv_None;
		}
		std::vector<XrdOucString>::const_iterator itr;
		for(itr=RedirConfig->AuthLibRestrict.begin();
				itr != RedirConfig->AuthLibRestrict.end();
				++itr) {
			if (fp.find(*itr) == 0) {
				restr = 0;
				break;
			}
		}
		if (restr) {
			TRACE(MOST,"Path vetoed, not in "
				"fixed id restrict list");
			return XrdAccPriv_None;
		}
		return ret_privs;
	}

	DEBUG("Passing for pure dpm authorization, "
		"proto=" << (Entity ? Entity->prot : "[none]"));
	
	return XrdAccPriv_All;
}

int XrdDPMRedirAcc::Test(const XrdAccPrivs priv,const Access_Operation oper)
{
	// Warning! This table must be in 1-to-1 correspondence with Access_Operation
	//
  	static XrdAccPrivs need[] = {XrdAccPriv_None,                 // 0
				XrdAccPriv_Chmod,                // 1
				XrdAccPriv_Chown,                // 2
				XrdAccPriv_Create,               // 3
				XrdAccPriv_Delete,               // 4
				XrdAccPriv_Insert,               // 5
				XrdAccPriv_Lock,                 // 6
				XrdAccPriv_Mkdir,                // 7
				XrdAccPriv_Read,                 // 8
				XrdAccPriv_Readdir,              // 9
				XrdAccPriv_Rename,               // 10
				XrdAccPriv_Lookup,               // 11
				XrdAccPriv_Update                // 12
                               
	};
	if (oper < 0 || oper > AOP_LastOp) return 0;
	return (int)(need[oper] & priv) == need[oper];
}

int XrdDPMRedirAcc::Audit(const int accok,
		const XrdSecEntity *Entity,
		const char *path,
		const Access_Operation oper,
		XrdOucEnv *Env)
{
	return accok;
}

extern "C"
{
XrdAccAuthorize *DpmXrdAccAuthorizeObject(XrdSysLogger *lp,
					const char   *cfn,
					const char   *parm,
					int ForceSecondary)
{
	EPNAME("DpmXrdAccAuthorizeObject");
	typedef XrdAccAuthorize* (*ept)(XrdSysLogger *, const char *, const char *);
	static bool initialised = 0;
	XrdSysPlugin *tokLib;

	if (!RedirConfig) {
		const char *p = getenv("XRDDPM_REDIRCONFIG");
		if (p) {
			void *p2;
			sscanf(p, "%p", &p2);
			RedirConfig = (DpmRedirConfigOptions*)p2;
		}
	}

	if (!initialised) {
		initialised = 1;

		Say.logger(lp);
		XrdSysError_Table *ETab = XrdDmliteError_Table();
		Say.addTable(ETab);

		XrdDmCommonInit();

		XrdOucString mypar(parm),AuthLib,Args;
		int from = 0;
		from = mypar.tokenize(AuthLib, from, ' ');
		if (from != STR_NPOS) { Args.assign(mypar, from, -1); }

		if (AuthLib.length()) {
			if (!(tokLib = new XrdSysPlugin(
					&Say, SafeCStr(AuthLib)))) {
				Say.Emsg("NewObject",
					"Could not open authorization library",
					SafeCStr(AuthLib));
				return 0;
			}
			ept ep = reinterpret_cast<ept>(reinterpret_cast<size_t>
				(tokLib->getPlugin("XrdAccAuthorizeObject")));
			if (!ep || (tokAuthorization =
					ep(lp, cfn, Args.length() ?
						Args.c_str() : 0)) == 0) {
				Say.Emsg("NewObject",
					"Could not get an authorization "
					"instance from libary",
					SafeCStr(AuthLib));
				delete tokLib;
				return 0;
			}
		}
	}

	if (ForceSecondary && !tokAuthorization) { return 0; }

	try {
		return new XrdDPMRedirAcc(cfn,ForceSecondary);
	} catch (dmlite::DmException &e) {
		Say.Emsg("NewObject","cannot start the access control layer",
			e.what());
	} catch(...) {
		Say.Emsg("NewObject","unexpected exception");
	}
	return 0;
}
} // extern "C"

extern "C"
{
XrdAccAuthorize *XrdAccAuthorizeObject(XrdSysLogger *lp,
                                       const char   *cfn,
                                       const char   *parm)
{
	EPNAME("XrdAccAuthorizeObject");

	return DpmXrdAccAuthorizeObject(lp,cfn,parm,0);
}
} // extern "C"
